#include "Integrable_Base.H"
#include "Beam_Spectra_Handler.H"
#include "ISR_Handler.H"

#include "Run_Parameter.H"
#include "Combined_Selector.H"
#include "Jet_Finder.H"
#include "Phase_Space_Handler.H"
#include "Regulator_Base.H"
#include "Running_AlphaS.H"
#include "Message.H"
#include "MyStrStream.H"

using namespace PHASIC;
using namespace MODEL;
using namespace ATOOLS;

#define CA 3.0
#define TR 0.5
#define Nf 5.0

Integrable_Base::Integrable_Base(const size_t nin,const size_t nout,
				 const scl::scheme scalescheme,const int kfactorscheme,
				 BEAM::Beam_Spectra_Handler *const beamhandler,
				 PDF::ISR_Handler *const isrhandler,
				 Selector_Data *const selectordata):
  m_name(""), m_nin(nin), m_nout(nout), m_naddin(0), m_naddout(0), 
  m_nvector(ATOOLS::Max(nin+nout,(size_t)1)), p_flavours(NULL), p_addflavours(NULL), 
  p_momenta(new Vec4D[ATOOLS::Max(nin+nout,(size_t)1)]), p_addmomenta(NULL), 
  m_scalescheme(scalescheme), m_kfactorscheme(kfactorscheme), 
  m_nstrong(0), m_neweak(0), m_orderQCD(-1), m_orderEW(-1), m_usepi(0),
  m_threshold(0.), m_overflow(0.), m_enhancefac(1.0), m_rfactor(1.0), m_xinfo(std::vector<double>(4)),
  m_n(0), m_expevents(1), m_dicedevents(0), m_accevents(0), m_last(0.), m_lastlumi(0.), m_lastdxs(0.), 
  m_max(0.), m_totalxs(0.),m_totalsum (0.), m_totalsumsqr(0.), m_totalerr(0.), 
  m_ssum(0.), m_ssumsqr(0.), m_smax(0.), m_ssigma2(0.), m_wmin(0.), m_me_as_factor(1.0), m_ycut(0.0), 
  m_sn(0), m_son(1), m_swaped(false), 
  p_selected(this), p_parent(this), 
  p_regulator(Regulator_Base::GetRegulator(this,"Identity",std::vector<double>())),
  p_beamhandler(beamhandler), p_isrhandler(isrhandler), 
  p_pshandler(NULL), p_activepshandler(NULL), p_selector(NULL), 
  p_cuts(NULL), p_whisto(NULL), p_jf(NULL), m_ownselector(true), m_efunc("1") 
{
  m_gmin=-1.0;
  SetScaleScheme(m_scalescheme);
  int kfs(ToType<int>(rpa.gen.Variable("S_KFACTOR_SCHEME","1"))&2);        
  double fssf(ToType<double>(rpa.gen.Variable("FS_CPL_SCALE_FACTOR","1.0")));
  double issf(ToType<double>(rpa.gen.Variable("IS_CPL_SCALE_FACTOR","1.0")));
  m_ps_kfactor=kfs?CA*(67./18.-M_PI*M_PI/6.)-10./9.*TR*Nf:0.0;
  m_ps_cpl_factor=Min(fssf,issf);
}

Integrable_Base::~Integrable_Base()
{
  if (p_selector!=NULL && m_ownselector) delete p_selector;
  if (p_flavours!=NULL) delete [] p_flavours;
  if (p_momenta!=NULL) delete [] p_momenta;
  if (p_addflavours!=NULL) delete [] p_addflavours;
  if (p_addmomenta!=NULL) delete [] p_addmomenta;
  if (p_whisto!=NULL) delete p_whisto;
}

Integrable_Base *const Integrable_Base::Selected()
{ 
  if (p_selected!=this && p_selected!=NULL) return p_selected->Selected();
  return this; 
}

Integrable_Base *const Integrable_Base::Parent()
{ 
  if (p_parent!=this && p_parent!=NULL) return p_parent->Parent();
  return this; 
}

double Integrable_Base::TotalResult()
{ 
//   if (m_ssigma2>0. && m_sn<1000) return m_totalsum/m_ssigma2; 
//   if (m_sn<1000) return m_ssum/m_sn;
//   double ssigma2 = (m_ssumsqr/m_sn - sqr(m_ssum/m_sn))/(m_sn-1);
//   return (m_totalsum+m_ssum/ssigma2/m_sn)/(m_ssigma2+1./ssigma2);
  if (m_ssigma2>0. && m_sn<1000) return m_totalsum/m_ssigma2; 
  if (m_sn<1000) return m_ssum/m_sn;
  double ssigma2 = sqr(m_ssum/m_sn)/((m_ssumsqr/m_sn - sqr(m_ssum/m_sn))/(m_sn-1));
  return (m_totalsum+m_ssum*ssigma2/m_sn)/(m_ssigma2+ssigma2);
}

double Integrable_Base::TotalVar() 
{
  if (m_nin==1 && m_nout==2) return 0.;
  if (m_sn<1000) {
    if (m_ssigma2>0.) return m_totalsum/m_ssigma2/sqrt(m_ssigma2); 
    else return TotalResult(); 
  }

  double disc = sqr(m_ssum/m_sn)/((m_ssumsqr/m_sn - sqr(m_ssum/m_sn))/(m_sn-1));
  if (disc>0.) return TotalResult()/sqrt(m_ssigma2+disc);
  
  return m_totalsum/m_ssigma2/sqrt(m_ssigma2);
}

double Integrable_Base::RemainTimeFactor(double maxerr) 
{
  if (m_sn<1000) return 0.;
  double disc = sqr(m_ssum/m_sn)/((m_ssumsqr/m_sn - sqr(m_ssum/m_sn))/(m_sn-1));
  return (sqr(1./maxerr)-m_ssigma2)/disc;
}

void Integrable_Base::SetMomenta(const Vec4D *momenta) 
{ 
  if (!p_momenta) {
    msg.Error()<<"Integrable_Base::SetMomenta("<<momenta<<"): "
		       <<"p_momenta = NULL. Abort."<<std::endl;
    abort();
  }
  for (size_t i=0;i<NVector();++i) p_momenta[i]=momenta[i];
  if (Selected()!=this) 
    for (size_t i=0;i<NVector();++i) 
      Selected()->p_momenta[i]=momenta[i];
}

void Integrable_Base::SetAddMomenta(const Vec4D *momenta) 
{ 
  if (!p_addmomenta) {
    msg.Error()<<"Integrable_Base::SetAddMomenta("<<momenta<<"): "
		       <<"p_addmomenta = NULL. Abort."<<std::endl;
    abort();
  }
  for (size_t i=0;i<NAddIn()+NAddOut();++i) p_addmomenta[i]=momenta[i];
  if (Selected()!=this) 
    for (size_t i=0;i<NAddIn()+NAddOut();++i) 
      Selected()->p_addmomenta[i]=momenta[i];
}

void Integrable_Base::SetAddFlavours(const Flavour *flavours) 
{ 
  if (!p_addflavours) {
    msg.Error()<<"Integrable_Base::SetAddFlavours("<<flavours<<"): "
		       <<"p_addflavours = NULL. Abort."<<std::endl;
    abort();
  }
  for (size_t i=0;i<NAddIn()+NAddOut();++i) p_addflavours[i]=flavours[i];
}

void Integrable_Base::InitWeightHistogram() 
{
  if (p_whisto!=NULL) {
    delete p_whisto; };
  double av=TotalResult();
  if (!av>0.) {
    msg.Error()<<"Integrable_Base::InitWeightHistogram(): "
		       <<"No valid result: "<<av<<std::endl;
    return;
  }
  if (av<.3) av/=10.;
  av = exp(log(10.)*int(log(av)/log(10.)+0.5));
  p_whisto = new Histogram(10,av*1.e-4,av*1.e6,100);
}

void Integrable_Base::ReadInHistogram(std::string dir)
{
  std::string filename = dir+"/"+m_name;
  std::ifstream from;
  bool     hit = 0;
  from.open(filename.c_str());
  if (from) hit = 1;
  from.close();
  if (!hit) return;
  if (p_whisto) delete p_whisto; 
  p_whisto = new Histogram(filename);	
}

void Integrable_Base::WriteOutHistogram(std::string dir)
{
  if (!p_whisto) return;
  p_whisto->Output(dir+"/"+m_name);	
}

double Integrable_Base::GetMaxEps(double epsilon)
{
  if (!p_whisto) return m_max;
  int ovn=int(p_whisto->Fills()*epsilon);
  if (ovn<1) return m_max;
  double maxeps = m_max;
  double min = m_max*epsilon; //upper boundary for weight reduction
  for (int i=p_whisto->Nbin()+1;i>0;i--) {
    if (ovn>=p_whisto->Value(i)) {
      ovn-=(int)p_whisto->Value(i);
      maxeps=exp(log(10.)*(p_whisto->Xmin()+(i-1)*p_whisto->BinSize()));
    }
    else return ATOOLS::Max(Min(maxeps,m_max),min);
  }
  return m_max;
}

void Integrable_Base::AddPoint(const double value) 
{
  m_n++;
  m_sn++;
  m_ssum    += value;
  m_ssumsqr += value*value;
  if (value>m_max)  m_max  = value;
  if (value>m_smax) m_smax = value;
  if (p_whisto) p_whisto->Insert(value);
}

void Integrable_Base::SetMax(const double max, int depth) 
{
  if (max!=0.) m_max=max;
} 

void Integrable_Base::SetMax() 
{
  msg.Error()<<"Integrable_Base::SetMax(): Virtual function called !"<<std::endl;
} 

void Integrable_Base::ResetMax(int) 
{
  msg.Error()<<"Integrable_Base::ResetMax(): Virtual function called !"<<std::endl;
} 

Blob_Data_Base *Integrable_Base::OneEvent() 
{
  msg.Error()<<"Integrable_Base::OneEvent(): Virtual function called !"<<std::endl;
  return false;
} 

#ifdef EVENT_STATISTICS 
#include <map>
#include <iomanip>

#define DFORMAT std::setw(15)
#define LIFORMAT std::setw(8)

typedef std::map<std::string,std::vector<double> > Mem_Map;

struct Memory_Map: public Mem_Map {

  // destructor
  ~Memory_Map();

};// end of struct Memory_Map

Memory_Map::~Memory_Map() 
{
  msg_Info()<<"Memory_Map::~Memory_Map(): Weighted analysis statistics {\n\n";
  {
    msg_Indent();
    std::cout.precision(6);
    double expectedxs=0.0, dicedxs=0.0;
    double expectedevents=0.0, dicedevents=0.0, untriggered=0.0;
    msg_Info()<<std::setw(25)<<std::setiosflags(std::ios::left)<<"Process"
	      <<std::resetiosflags(std::ios::left)<<std::setiosflags(std::ios::right)
	      <<DFORMAT<<"exp. xs"<<DFORMAT<<"diced xs"
	      <<LIFORMAT<<"exp. n"<<LIFORMAT<<"diced n"
	      <<LIFORMAT<<"acc. n"<<LIFORMAT<<"extra n"
	      <<std::resetiosflags(std::ios::right)<<std::endl;
    for (Mem_Map::const_iterator mit=begin();mit!=end();++mit) {
      if (mit->second.size()==0) return;
      expectedxs+=mit->second[0];
      expectedevents+=mit->second[1];
      dicedxs+=mit->second[2];
      dicedevents+=mit->second[3];
      untriggered+=mit->second[4];
      msg_Info()<<std::setw(25)<<std::setiosflags(std::ios::left)<<mit->first
		<<std::resetiosflags(std::ios::left)<<std::setiosflags(std::ios::right)
		<<DFORMAT<<(mit->second[0]*rpa.Picobarn())
		<<DFORMAT<<(mit->second[2]/mit->second[4]*rpa.Picobarn())
		<<LIFORMAT<<mit->second[1]<<LIFORMAT<<mit->second[4]
		<<LIFORMAT<<mit->second[3]<<LIFORMAT<<mit->second[5]
		<<std::resetiosflags(std::ios::right)<<std::endl;
    }
    msg_Info()<<"\n";
    msg_Info()<<"exp. n = "<<expectedevents<<", diced n = "<<untriggered
	      <<", acc. n = "<<dicedevents<<"\n";
    msg_Info()<<"exp. xs = "<<(expectedxs/untriggered*rpa.Picobarn())
	      <<", diced xs = "<<(dicedxs/untriggered*rpa.Picobarn())<<std::endl;
  }
  msg_Info()<<"\n}"<<std::endl;
}

static Memory_Map mem;
#endif

bool Integrable_Base::Trigger(const Vec4D *const momenta)
{
  return p_selector->Trigger(momenta);
} 

Blob_Data_Base *Integrable_Base::OneEvent(const double mass,const int mode) 
{
  return p_activepshandler->OneEvent(mass,mode);
} 

Blob_Data_Base *Integrable_Base::SameEvent() 
{
  return p_activepshandler->SameEvent();
  msg.Error()<<"Integrable_Base::SameEvent(): Virtual function called !"<<std::endl;
  return false;
} 

Blob_Data_Base *Integrable_Base::WeightedEvent(const int mode) 
{
  msg.Error()<<"Integrable_Base::WeightedEvent(): Virtual function called !"<<std::endl;
  return NULL;
} 

Blob_Data_Base *Integrable_Base::WeightedEventNS(const int mode) 
{
  msg.Error()<<"Integrable_Base::WeightedEventNS(): Virtual function called !"<<std::endl;
  return NULL;
} 

Blob_Data_Base *Integrable_Base::SameWeightedEvent() 
{
  msg.Error()<<"Integrable_Base::SameWeightedEvent(): Virtual function called !"<<std::endl;
  return NULL;
} 

void Integrable_Base::SetPSHandler(Phase_Space_Handler *const pshandler) 
{
  p_activepshandler=pshandler;
} 

void Integrable_Base::OptimizeResult()
{
  msg.Error()<<"Integrable_Base::OptimizeResult(): Virtual function called !"<<std::endl;
} 

void Integrable_Base::SetMomenta()   
{ 
  const Vec4D *p=p_activepshandler->Point();
  for (size_t i=0;i<m_nvector;++i) p_momenta[i]=p[i];
}

double Integrable_Base::CalculateScale(const Vec4D *momenta) 
{
  if (!m_kfkey.Assigned()) {
    std::string kfname(p_activepshandler->Process()->Name());
    std::string kfinfo("O(QCD)="+ToString(m_orderQCD));
    msg_Debugging()<<METHOD<<"(): Assign '"<<Name()
		   <<"' to '"<<kfname<<"','"<<kfinfo<<"'\n";
    m_kfkey.Assign(kfname,2,0,p_activepshandler->GetInfo());
    m_kfkey.SetInfo(kfinfo);
  }
  scl::scheme scheme(m_scalescheme);
  if (scheme==scl::unknown) 
    THROW(fatal_error,"Unknown scale scheme: "+ToString(m_scalescheme));
  SetMomenta(momenta);
  if (m_nin==1) return momenta[0].Abs2();
  if (m_nin!=2) THROW(fatal_error,"Too many incoming particles.");
  if (scheme==scl::ckkw) {
    double S(sqr(rpa.gen.Ecms()));
    m_scale[stp::fac]=rpa.gen.Ycut()*S;
    m_scale[stp::ren]=sqr(rpa.gen.DeltaR())*m_ycut*S;
    if (p_jf==NULL && p_selector->Name()=="Combined_Selector") {
      p_jf=(Jet_Finder *)
	((Combined_Selector*)p_selector)->GetSelector("Jetfinder");
      if (p_jf!=NULL) m_me_as_factor=p_jf->Type()>1?1.0:0.25;
      else THROW(critical_error,"'SCALE_SCHEME = CKKW' implies JetFinder <ycut> <deltar>' in 'Selector.dat'.");
    }
    double pt2(p_jf->ActualValue()*S);
    if ((int)m_nout==m_maxjetnumber) 
      // highest multiplicity treatment
      if (m_nstrong>2) m_scale[stp::fac]=pt2;
    else 
      // two scale treatment
      m_scale[stp::fac]=Min(m_scale[stp::fac],pt2);
    m_kfkey[0]=m_scale[stp::ren]*=rpa.gen.RenormalizationScaleFactor();
    m_kfkey[1]=m_scale[stp::fac]*=rpa.gen.FactorizationScaleFactor();
    return m_scale[stp::fac]*m_ps_cpl_factor;
  }
  double pt2(-1.0);
  if (scheme & scl::fixed) {
    pt2=m_scale[stp::fac];
    scheme-=scl::fixed;
  }
  else if (scheme & scl::shat) {
    pt2=(momenta[0]+momenta[1]).Abs2();
    scheme-=scl::shat;
  }
  else if (scheme & scl::gmeanpt) {
    if (m_nin+m_nout==4) {
      double s = (momenta[0]+momenta[1]).Abs2();
      double t = (momenta[0]-momenta[2]).Abs2()-
	(sqr(p_flavours[2].PSMass())+sqr(p_flavours[3].PSMass()))/2.;
      double u = (momenta[0]-momenta[3]).Abs2()-
	(sqr(p_flavours[2].PSMass())+sqr(p_flavours[3].PSMass()))/2.;
      pt2 = 2.*s*t*u/(s*s+t*t+u*u);
    }
    else {
      pt2 = 1.;
      int count(0);
      for (size_t i=m_nin;i<m_nin+m_nout;i++) {
	if ((scheme & scl::strong) && !p_flavours[i].Strong()) continue;
	if (scheme & scl::mass) pt2 *= momenta[i].MPerp2();
	else pt2 *= momenta[i].PPerp2();
	count++;
      }
      if (count==0) pt2=(momenta[0]+momenta[1]).Abs2();
      else pt2 = pow(pt2,1./count);
    }
    scheme-=scl::gmeanpt;
    scheme-=scl::strong;
    scheme-=scl::mass;
  }
  else if (scheme & scl::ameanpt) {
    pt2 = 0.;
    int count(0);
    for (size_t i=m_nin;i<m_nin+m_nout;i++) {
      if ((scheme & scl::strong) && !p_flavours[i].Strong()) continue;
      if (scheme & scl::mass) pt2 += momenta[i].MPerp2();
      else pt2 += momenta[i].PPerp2();
      count++;
    }
    if (count==0) pt2=(momenta[0]+momenta[1]).Abs2();
    else pt2/=count;
    scheme-=scl::ameanpt;
    scheme-=scl::strong;
    scheme-=scl::mass;
  }
  else if (scheme & scl::minpt) {
    pt2 = (momenta[0]+momenta[1]).Abs2();
    for (size_t i=m_nin;i<m_nin+m_nout;i++) {
      if ((scheme & scl::strong) && !p_flavours[i].Strong()) continue;
      if (scheme & scl::mass) pt2 = Min(pt2,momenta[i].MPerp2());
      else pt2 = Min(pt2,momenta[i].PPerp2());
    }
    scheme-=scl::minpt;
    scheme-=scl::strong;
    scheme-=scl::mass;
  }
  else if (scheme & scl::sumpt) {
    pt2 = 0.;
    for (size_t i=m_nin;i<m_nin+m_nout;i++) {
      if ((scheme & scl::strong) && !p_flavours[i].Strong()) continue;
      if (scheme & scl::mass) pt2 += momenta[i].MPerp2();
      else pt2 += momenta[i].PPerp2();
    }
    scheme-=scl::sumpt;
    scheme-=scl::strong;
    scheme-=scl::mass;
  }
  if (scheme & scl::div_by_2) { 
    pt2/=2.; 
    scheme-=scl::div_by_2; 
  }
  if (scheme & scl::mult_by_2) { 
    pt2*=2.; 
    scheme-=scl::mult_by_2;  
  }
  if (pt2<0. || scheme!=scl::unknown) 
    THROW(fatal_error,"Unknown scale scheme: "+ToString(m_scalescheme));
  m_kfkey[0]=m_scale[stp::ren]=rpa.gen.RenormalizationScaleFactor()*pt2;
  m_kfkey[1]=m_scale[stp::fac]=rpa.gen.FactorizationScaleFactor()*pt2;
  if (Selected()==NULL) return m_scale[stp::fac]*m_ps_cpl_factor;
  return (*Selected()->p_regulator)[m_scale[stp::fac]*m_ps_cpl_factor]; 
}

double Integrable_Base::KFactor() 
{
  if (!m_kfkey.Assigned()) {
    std::string kfname(p_activepshandler->Process()->Name());
    std::string kfinfo("O(QCD)="+ToString(m_orderQCD));
    msg_Debugging()<<METHOD<<"(): Assign '"<<Name()
		   <<"' to '"<<kfname<<"','"<<kfinfo<<"'\n";
    m_kfkey.Assign(kfname,2,0,p_activepshandler->GetInfo());
    m_kfkey.SetInfo(kfinfo);
  }
  if (m_scalescheme&scl::ckkw) {
    if (m_kfkey.Weight()!=ATOOLS::UNDEFINED_WEIGHT) return m_kfkey.Weight();
    if (m_orderQCD<0 || m_orderEW<0) {
      THROW(fatal_error,"Couplings not set for process '"+Name()+"'");
    }
    if (m_nstrong<=2) return m_rfactor;
    m_scale[stp::ren]=m_kfkey[0];
    m_scale[stp::fac]=m_kfkey[1];
    double asn(as->AlphaS(m_me_as_factor*m_ps_cpl_factor*m_scale[stp::ren]));
    if (m_ps_kfactor!=0.0) asn*=1.+asn/(2.0*M_PI)*m_ps_kfactor;
    m_kfkey<<m_rfactor*pow(asn/as->AlphaS(sqr(rpa.gen.Ecms())),m_orderQCD);
    msg_Debugging()<<METHOD<<"(): "<<Name()<<" ("<<m_nstrong<<","
		   <<m_orderQCD<<") {\n"
		   <<"  \\mu_{fac}   = "<<sqrt(m_scale[stp::fac])<<"\n"
		   <<"  \\mu_{ren}   = "<<sqrt(m_scale[stp::ren])<<"\n"
		   <<"  me scalefac = "<<m_me_as_factor<<"\n"
		   <<"  r scalefac  = "<<m_rfactor<<"\n"
		   <<"  ps scalefac = "<<m_ps_cpl_factor<<"\n"
		   <<"  ps k factor = "<<m_ps_kfactor
		   <<"\n} -> as = "<<asn<<" => K = "<<m_kfkey.Weight()<<"\n";
    return m_kfkey.Weight();
  }
  if (m_kfactorscheme==1) {
    if (m_kfkey.Weight()!=ATOOLS::UNDEFINED_WEIGHT) return m_kfkey.Weight();
    if (m_orderQCD<0 || m_orderEW<0) {
      THROW(fatal_error,"Couplings not set for process '"+Name()+"'");
    }
    if (m_orderQCD>0) {
      m_scale[stp::ren]=m_kfkey[0];
      m_kfkey<<m_rfactor*pow(as->AlphaS(m_scale[stp::ren])/
			     as->AlphaS(sqr(rpa.gen.Ecms())),m_orderQCD);
      return m_kfkey.Weight();
    } 
    else 
      return m_rfactor;
  }
  return 1.;
}

void Integrable_Base::SetEvents(const double number)
{
  m_anasum=m_validanasum=0.0;
  m_gmin=1.;
  m_expevents=(long int)(0.6*number*TotalXS()/TriggerEfficiency()+1.0);
  m_dicedevents=m_accevents=0;
#ifdef EVENT_STATISTICS  
  std::vector<double> smem(6,0.0);
  smem[0]=TotalResult();
  smem[1]=m_expevents;
  mem[Name()]=smem;
#endif
}

void Integrable_Base::AddEvent(const double xs,const double validxs,
			       const int ncounts)
{
  m_anasum+=xs;
  m_validanasum+=validxs;
  m_dicedevents+=ncounts; 
  if (xs!=0.0) m_accevents++;
#ifdef EVENT_STATISTICS  
  if (mem[Name()].size()==0) {
    std::vector<double> smem(6,0.0);
    smem[0]=TotalResult();
    smem[1]=m_expevents;
    mem[Name()]=smem;
  }
  mem[Name()][4]+=ncounts;
  mem[Name()][2]+=xs;
  if (xs!=0.0) mem[Name()][3]++;
#endif
}

double Integrable_Base::TriggerEfficiency()
{
  if (p_pshandler!=NULL) {
    Multi_Channel *fsr=p_pshandler->FSRIntegrator();
    return fsr->ValidN()/(double)fsr->N();
  }
  return 1.0;
}

void Integrable_Base::ResetEvents(double gmin)
{
  m_gmin = gmin;
#ifdef EVENT_STATISTICS  
  mem[Name()][5]=m_expevents-mem[Name()][1];
#endif
  if (m_validanasum==0.) return;
  long int expevents=m_expevents;
  m_expevents=(long int)(m_anasum/m_validanasum/gmin*(double)m_expevents+.5);
  m_expevents=(long int)Min((double)m_expevents,expevents*sqrt(expevents)/gmin);
  if (!(m_expevents>=expevents)) m_expevents=expevents;
#ifdef EVENT_STATISTICS  
  mem[Name()][5]=m_expevents-expevents;
  msg_Tracking()<<Name()<<" -> extra events "<<expevents<<" "
		<<m_expevents<<" "<<mem[Name()][5]<<" "
		<<m_anasum/m_validanasum<<" "<<gmin<<std::endl;
#endif
}

void Integrable_Base::GetGMin(double &g, double &meff)
{
  if (m_validanasum==0.) return;
  double q = m_anasum/m_validanasum;
  double add = q*(double)m_dicedevents;
  add = Min(add,sqrt((double)m_expevents)*(double)m_dicedevents);
  meff += add;
  if (m_expevents<100) return;
  if (q<1.) if (m_expevents<sqrt(Parent()->m_expevents)) q=1.;
  g = Min(g,q);
}

void Integrable_Base::SetISRThreshold(const double threshold) 
{
  m_threshold=threshold;
}

void Integrable_Base::CreateMomenta(const size_t n)
{ 
  delete [] p_momenta; 
  p_momenta = new Vec4D[n]; 
  m_nvector=n; 
}

Spin_Correlation_Tensor* Integrable_Base::GetSpinCorrelations()
{
  return NULL;
}

void Integrable_Base::FillAmplitudes(ATOOLS::Amplitude_Tensor*,double)
{
  return;
}

std::map<std::string,std::string> Integrable_Base::ScaleTags()
{
  std::map<std::string,std::string> tags;
  tags["UNKNOWN"]=ToString(scl::unknown);
  tags["CKKW"]=ToString(scl::ckkw);
  tags["FIX_SCALE"]=ToString(scl::fixed);
  tags["S_HAT"]=ToString(scl::shat);
  tags["A_MEAN_PT2"]=ToString(scl::ameanpt);
  tags["G_MEAN_PT2"]=ToString(scl::gmeanpt);
  tags["MIN_PT2"]=ToString(scl::minpt);
  tags["SUM_PT2"]=ToString(scl::sumpt);
  tags["STRONG"]=ToString(scl::strong);
  tags["MASSIVE"]=ToString(scl::mass);
  tags["HALVE"]=ToString(scl::div_by_2);
  tags["DOUBLE"]=ToString(scl::mult_by_2);
  tags["UPDF"]=ToString(scl::updf);
  return tags;
}
 
void Integrable_Base::SetScaleScheme(const scl::scheme s)
{ 
  m_scalescheme=s;   
  if ((m_scalescheme&scl::ckkw) && m_ycut<=0.0) {
    m_ycut=rpa.gen.Ycut();
    m_scale[stp::fac]=m_ycut*sqr(rpa.gen.Ecms())*
      rpa.gen.FactorizationScaleFactor();
    m_scale[stp::ren]=m_ycut*sqr(rpa.gen.DeltaR()*rpa.gen.Ecms())*
      rpa.gen.RenormalizationScaleFactor();
  }
}
